/* 
 * Exemple d'utilisation du stencil buffer : effet miroir.
 *
 * Auteur : Fares BELHADJ (Lire COPYING)
 * Email  : amsi@ai.univ-paris8.fr
 * Date   : 17/01/2008
 */

#include <math.h>
#include <stdio.h>
#include <assert.h>
#include <stdlib.h>
#include <SDL.h>
#include <GL/gl.h>
#include <GL/glu.h>

#define INF 1000000

/*
 * Prototypes des fonctions statiques.
 */
static SDL_Surface * sdlInit        (int w, int h);
static void          sdlInitGL      (SDL_Surface * s);
static void          sdlResizeGL    (SDL_Surface * s);
static void          sdlLoop        (SDL_Surface * s);
static int           sdlManageEvents(SDL_Surface * s);
static void          sdlDraw        (SDL_Surface * s);
static void          scene          (void);

/*
 * coordonnees de souris et position de la lumiere.
 */
static int drag_x = -INF, drag_y = -INF, drag_b = -INF;
static GLfloat lightpos[] = { -5.0, 1.0, 5.0, 0.0 };

/*
 * pour translater la sphere.
 */
static GLfloat dtx = 0, dty = 0, dtz = 0, tx = 0, ty = 0, tz = 2;

/* un tout petit main */
int main(int argc, char ** argv) {
  SDL_Surface * s;
  s = sdlInit(640, 480);
  sdlLoop(s);
  return 0;
}

/* sdlInit : initialiser SDL pour OpenGL */
static SDL_Surface * sdlInit(int w, int h) {
  SDL_Surface * s;
  const SDL_VideoInfo * vi = NULL;
  if ( SDL_Init(SDL_INIT_VIDEO) < 0 ) {
    fprintf(stderr, "Impossible d'initialiser SDL: %s\n", SDL_GetError());
    exit(1);
  }

  if( (vi = SDL_GetVideoInfo()) == 0 ) {
    fprintf(stderr, "Echec de SDL_GetVideoInfo(): %s\n", SDL_GetError());
    exit(1);
  }
  SDL_GL_SetAttribute( SDL_GL_RED_SIZE,     5 );
  SDL_GL_SetAttribute( SDL_GL_GREEN_SIZE,   5 );
  SDL_GL_SetAttribute( SDL_GL_BLUE_SIZE,    5 );
  SDL_GL_SetAttribute( SDL_GL_DEPTH_SIZE,   16 );
  SDL_GL_SetAttribute( SDL_GL_DOUBLEBUFFER, 1 );
  SDL_GL_SetAttribute( SDL_GL_STENCIL_SIZE, 1 ); /* Nouveau !!! active le mode stencil */

  if ( (s = SDL_SetVideoMode(w, h, vi->vfmt->BitsPerPixel,
			     SDL_OPENGL | SDL_RESIZABLE)) == NULL ) {
    fprintf(stderr, "Impossible d'ouvrir le mode video : %s\n", SDL_GetError());
    exit(1);
  }
  atexit(SDL_Quit);
  sdlInitGL(s);
  return s;
}

/* 
 * sdlInitGL : configurer les options GL.
 */
static void sdlInitGL(SDL_Surface * s) {
  GLfloat color[] = { 1.0, 1.0, 1.0, 1.0 };
  glShadeModel(GL_SMOOTH);
  glDepthFunc(GL_LEQUAL);
  glClearColor(0.1, 0.1, 0.1, 1);
  glClearDepth(1.0);
  glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);
  glDisable(GL_TEXTURE_2D);
  glEnable(GL_CULL_FACE);
  glEnable(GL_DEPTH_TEST);
  glClearStencil(0.0); /* Nouveau !!! la valeur d'effacement du stencil */

  /* parametrer la lumiere ... */
  glLightfv(GL_LIGHT0, GL_AMBIENT, color);
  glLightfv(GL_LIGHT0, GL_DIFFUSE, color);
  glLightfv(GL_LIGHT0, GL_SPECULAR, color);
  glEnable(GL_LIGHTING);
  glEnable(GL_LIGHT0); 

  sdlResizeGL(s);
}

/* 
 * CallBack : en cas de resize modifier la matrice de projection en
 * fonction du rapport w/h
 */
static void sdlResizeGL(SDL_Surface * s) {
  glViewport(0, 0, s->w, s->h);
  glMatrixMode(GL_PROJECTION);
  glLoadIdentity();
  gluPerspective(45.0, s->w / (float)s->h, 0.1, 100);
  glMatrixMode( GL_MODELVIEW );
  glLoadIdentity( );
}

/*
 * loop principale.
 */
static void sdlLoop(SDL_Surface * s) {
  for(;;) {
    while(sdlManageEvents(s)) 
      SDL_Delay(1);
    sdlDraw(s);
  }
}

/* 
 * sdlManageEvents gere les evenements ; aussi renvoie vrai si la
 * fenetre est inactive, sinon faux.
 * Les evenements sont : 
 *
 * La souris
 *
 * escape, q, Q : quitter
 *
 * w : activer / desactiver le mode wireframe
 *
 * l : activer / desactiver l'eclairage
 *
 * F1 : fullscreen On / Off
 * 
 * redimensionnement de fenetre : appeler sdlResizeGL
 *
 */
static int sdlManageEvents(SDL_Surface * s) {
  GLint v[2];
  static int active = 1;
  SDL_Event event;
  while(SDL_PollEvent(&event)) 
    switch (event.type) {
    case SDL_MOUSEBUTTONDOWN:
      drag_b = event.button.button; 
      drag_x = event.button.x;
      drag_y = event.button.y;
      break;
    case SDL_MOUSEBUTTONUP:
      drag_b = drag_x = drag_y = -INF;
      tx += dtx; dtx = 0; ty += dty; dty = 0; tz += dtz; dtz = 0;
      break;
    case SDL_MOUSEMOTION:
      if(drag_x != -INF && drag_y != -INF) {
	if(drag_b == SDL_BUTTON_LEFT) {
	  dtx = -4.0 * (drag_x - event.motion.x) / (float)(s->w);
	  dty =  4.0 * (drag_y - event.motion.y) / (float)(s->h);
	} else if(drag_b == SDL_BUTTON_MIDDLE) {
	  dtz = -4.0 * (drag_x - event.motion.x) / (float)(s->w);
	  dty =  4.0 * (drag_y - event.motion.y) / (float)(s->h);
	}
      }
      break;
    case SDL_KEYDOWN:
      switch(event.key.keysym.sym) {
      case 'w':
	glGetIntegerv(GL_POLYGON_MODE, v);
	if(v[0] == GL_FILL)
	  glPolygonMode(GL_FRONT, GL_LINE);
	else
	  glPolygonMode(GL_FRONT, GL_FILL);
	break;
      case 'l':
	if(glIsEnabled(GL_LIGHTING)) {
	  glDisable(GL_LIGHTING);
	  glDisable(GL_LIGHT0);
	} else {
	  glEnable(GL_LIGHTING);
	  glEnable(GL_LIGHT0);
	}
	break;
      case SDLK_ESCAPE:
      case 'q':
	exit(0);
      case SDLK_F1:
	SDL_WM_ToggleFullScreen(s);
	break;
      default:
	fprintf(stderr, "La touche %s a ete pressee\n",
		SDL_GetKeyName(event.key.keysym.sym));
	break;
      }
      break;
    case SDL_KEYUP:
      break;
    case SDL_ACTIVEEVENT:
      if(event.active.state & SDL_APPACTIVE)
	active = event.active.gain;
      break;
    case SDL_VIDEORESIZE:
      if( (s = SDL_SetVideoMode(event.resize.w, event.resize.h, 
				s->format->BitsPerPixel, s->flags)) == NULL ) {
	fprintf(stderr, "Echec de SDL_SetVideoMode apres le resize : %s\n", 
		SDL_GetError());
	exit(1);
      }
      sdlResizeGL(s);
      break;
    case SDL_QUIT:
      exit(0);
    }
  return !active;
}

/*
 * dessiner la scene ...
 */
static void sdlDraw(SDL_Surface * s) {
  /* Nouveau !!! effacer le stencil avec les autres buffers */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT | GL_STENCIL_BUFFER_BIT);
  glPushMatrix();
  glLoadIdentity();
  gluLookAt(-5, 0, 5, 0, 0, 0, 0, 1, 0);
  /* lumiere positionnelle */
  glLightfv(GL_LIGHT0, GL_POSITION, lightpos);
  /* la fonction scene initialise une seule fois la scene (genere les
     listes) puis appelle les objets de la scene afin */
  scene();
  glPopMatrix();
  SDL_GL_SwapBuffers();
}

/*
 * ... la scene.
 */
static void scene(void) {
  GLdouble a[4] = {0, 0, -1, 0};
  static GLuint obj = 0, plan = 0;
  if(!obj) { /* n'est appelee qu'une seule fois et permet de creer les
		listes d'affichage */
    GLUquadricObj * q;
    GLfloat color[4] = {0.25, 0.25, 0.5, 0.5};
    q = gluNewQuadric();
    assert(q);
    gluQuadricNormals(q, GLU_SMOOTH); /* smooth est par defaut */
    gluQuadricTexture(q, GL_FALSE); /* false est par defaut */
    
    /* ------------ Le miroir ------------ */
    plan = glGenLists(1);
    glNewList(plan, GL_COMPILE);

    glPushAttrib(GL_ALL_ATTRIB_BITS); /* sauvegarde tous les attributs (les enable/disable) */
    glEnable(GL_BLEND); /* active le blending (la gestion de transparence) */
    /* glBlendFunc(source_factor, destination_factor); determine le
       calcul de la transparence (quelles ponderations pour le melange
       de couleurs) */
    glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);
    
    /* du cuivre ... */
    glMaterialfv(GL_FRONT, GL_AMBIENT, color); glColor4fv(color);
    color[0] = 0.400;     color[1] = 0.400;     color[2] = 0.400;
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
    color[0] =  0.774597; color[1] =  0.774597; color[2] =  0.774597;
    glMaterialfv(GL_FRONT, GL_SPECULAR, color);
    color[0] = 0.6;
    glMaterialf(GL_FRONT, GL_SHININESS, color[0] * 128);

    /* un quad pour le miroir */
    glPushMatrix();
    glScalef(3, 3, 3);
    glBegin(GL_QUADS);
    glNormal3f(0, 0, 1);
    glVertex3f(-1, -1, 0);
    glVertex3f( 1, -1, 0);
    glVertex3f( 1,  1, 0);
    glVertex3f(-1,  1, 0);
    glEnd();
    glPopMatrix();
    glPopAttrib();
    glEndList();

    /* ------------ L'objet devant le miroir (une sphere) ------------ */
    obj = glGenLists(1);
    glNewList(obj, GL_COMPILE);

    /* du chrome bleute svp ... */
    color[0] = 0.2125;   color[1] = 0.1275;   color[2] = 0.054;   color[3] = 1.0;
    glMaterialfv(GL_FRONT, GL_AMBIENT, color); glColor4fv(color);
    color[0] = 0.714;    color[1] = 0.4284;   color[2] = 0.18144;
    glMaterialfv(GL_FRONT, GL_DIFFUSE, color);
    color[0] = 0.393548; color[1] = 0.271906; color[2] = 0.166721;
    glMaterialfv(GL_FRONT, GL_SPECULAR, color);
    color[0] = 0.2;
    glMaterialf(GL_FRONT, GL_SHININESS, color[0] * 128);

    gluSphere(q, 0.4, 15, 15);
    glEndList();
    
    gluDeleteQuadric(q);
    /* ca y est ... j'ai ce qu'il me faut ! */
  }

  /************ PARTIE APPELEE A CHAQUE AFFICHAGE *************/

  glEnable(GL_STENCIL_TEST); /* Nouveau !!! activer le test stencil */
  /* glStencilFunc(func, ref, mask); determine la fonction de test */
  glStencilFunc(GL_ALWAYS, 1, 1);
  /* glStencilOp (fail, failz, zpass); determine les actions realisees
     en fonction du resultat obtenu par le test */
  glStencilOp  (GL_REPLACE, GL_REPLACE, GL_REPLACE);

  glPushAttrib(GL_ALL_ATTRIB_BITS);
  /* toujours rendre (ici dans le stencil) le miroir en mode plein
     pour que tous ces fragments soient marques a 1 dans le stencil */
  glPolygonMode(GL_FRONT, GL_FILL);
  glPushMatrix();
  glCallList(plan);
  glPopMatrix();
  glPopAttrib();

  /* on efface le color et le depth, on garde que le stencil */
  glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
  /* ne faire passer que les fragments precedemment marques a 1 */
  glStencilFunc(GL_EQUAL, 1, 1);
  glStencilOp  (GL_KEEP, GL_KEEP, GL_KEEP);

  /* ---------- voici le reflet ... ---------- */
  glPushMatrix();
  /* Nouveau !!! plan de clipping (coupe) */
  glEnable(GL_CLIP_PLANE0); /* activer un plan de coupe */
  glClipPlane(GL_CLIP_PLANE0, a); /* "a[0 ... 3]" sont les coefs du plan de coupe */
  glScalef(1, 1, -1); /* miroir en zzz */
  glTranslatef(tx + dtx, ty + dty, tz + dtz); /* translater le reflet */
  glCallList(obj); /* l'objet dans la glace */
  glDisable(GL_CLIP_PLANE0);/* desactiver le plan de coupe */
  glPopMatrix();
  glDisable(GL_STENCIL_TEST); /* nous en avons fini avec le stencil */

  /* ---------- ... miroir mon beau miroir ... ---------- */
  glPushMatrix();
  glCallList(plan);
  glPopMatrix();

  /* ---------- ... la mechante reine ... ---------- */
  glPushMatrix();
  glTranslatef(tx + dtx, ty + dty, tz + dtz);
  glCallList(obj);
  glPopMatrix();
}

